﻿-- There is no copyright on this code

-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
-- associated documentation files (the "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is furnished to do so.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
-- NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

--
-- Global variables
--
local addon, HowMuch = ...
HowMuch.name = "HowMuch"
HowMuch.slashName = "hm"
HowMuch.version = "v1.3"
HowMuch.lang = "English"
HowMuch.startCount = 1
HowMuch.startup = false
HowMuch.batchSize = 50
HowMuch.datesToKeep = 14
HowMuch.db = {}
HowMuch.derivedDb = {}

local function SetPhaseDialog(phaseNumber, param1, param2)
	for i=1,6 do
		local item = "text" .. i
		if phaseNumber == i then
			HowMuch.window[item]:SetFontColor(1, 1, 1, 1)
		else
			HowMuch.window[item]:SetFontColor(132 / 255, 132 / 255, 132 / 255, 1)
		end
	end
	
	if phaseNumber == 1 then
		HowMuch.window.text1:SetText("Requesting list of all auctions from the server")
		HowMuch.window.text2:SetText("Waiting on auction list (takes a while)...")
		HowMuch.window.text3:SetText("Received auction list")
		HowMuch.window.text4:SetText("Requesting details on each auction from the server")
		HowMuch.window.text5:SetText("Received auction details")
		HowMuch.window.text6:SetText("Done")
	elseif phaseNumber == 3 then
		HowMuch.window.text3:SetText("Received list of " .. tostring(param1) .. " auctions")
	elseif phaseNumber == 5 then
		HowMuch.window.text5:SetText("Received details on " .. tostring(param1) .. " out of " .. tostring(param2) .. " auctions")
	end
end

local function GetCurrencyValues(ccyValue)
	local plat = math.floor(ccyValue / 10000)
	local remainder = ccyValue - plat * 10000
	local gold = math.floor(remainder / 100)
	local silver = remainder - gold * 100
	return plat, gold, silver
end

local function FormatCurrencyString(ccyValue)
	local plat, gold, silver = GetCurrencyValues(ccyValue)
	if plat > 0 then
		return plat .. "p " .. gold .. "g " .. silver .. "s"
	elseif gold > 0 then
		return gold .. "g " .. silver .. "s"
	else
		return silver .. "s"
	end
end

local function GetMedian(sortedArray)
	if #sortedArray == 0 then
		return 0
	end
	
	if #sortedArray == 1 then
		return sortedArray[1]
	end
	
	if math.fmod(#sortedArray,2) == 0 then
		local val1 = sortedArray[#sortedArray/2]
		local val2 = sortedArray[(#sortedArray/2)+1]
		if val1 == nil or val2 == nil then
			--print ("Len=" .. #sortedArray .. " val1=" .. tostring(val1) .. " val2=" .. tostring(val2))
			return 0
		else
			return ( sortedArray[#sortedArray/2] + sortedArray[(#sortedArray/2)+1] ) / 2
		end
	else
		return sortedArray[math.ceil(#sortedArray/2)]
	end
end

local function GetTrimmedAverage(sortedArray)
	if #sortedArray == 0 then
		return 0
	end
	
	if #sortedArray == 1 then
		return sortedArray[1]
	end
	
	if #sortedArray == 2 then
		return (sortedArray[1] + sortedArray[2]) / 2
	end
	
	-- Trimming outliers via quartiles.  See http://en.wikipedia.org/wiki/Quartile
	local lowerQuartileData = {}
	local upperQuartileData = {}
	local medianIndex = math.floor((#sortedArray/2.0 + 0.5) * 10) / 10
	
	for index, value in ipairs(sortedArray) do
		if index < medianIndex then
			table.insert(lowerQuartileData, value)
		elseif index > medianIndex then
			table.insert(upperQuartileData, value)
		end
	end
	
	--if math.fmod(#sortedArray,2) ~= 0 then
	--	table.insert(lowerQuartileData, sortedArray[medianIndex])
	--end
	
	--if #lowerQuartileData ~= #upperQuartileData then
		--print ("quartiles differ in size. total=" .. #sortedArray .. " lower=" .. #lowerQuartileData .. " upper=" .. #upperQuartileData)
	--end
	
	local lowerQuartileMedian = GetMedian(lowerQuartileData)
	local upperQuartileMedian = GetMedian(upperQuartileData)
	local interQuartileRange = upperQuartileMedian - lowerQuartileMedian
	local lowerBound = lowerQuartileMedian - (1.5 * interQuartileRange)
	local upperBound = upperQuartileMedian + (1.5 * interQuartileRange)
	local total = 0
	local count = 0
	for _, value in ipairs(sortedArray) do
		if value >= lowerBound and value <= upperBound then
			total = total + value
			count = count + 1
		end
	end
	
	if count < 1 then
		return 0
	else
		return math.ceil(total / count)
	end
end

local function GetSortedValues(item)
	local itemValues = {}
	local minValue = 999999999
	local maxValue = 0
	for _, itemData in pairs(item) do
		for price, data in pairs(itemData) do
			if price < minValue then
				minValue = price
			end
			
			if price > maxValue then
				maxValue = price
			end
			
			for i=0,data.count do
				table.insert(itemValues, price)
			end
		end
	end
	
	table.sort(itemValues)
	return minValue, maxValue, itemValues
end

local function GetAverageValues(item)
	local minValue, maxValue, sortedArray = GetSortedValues(item)
	local average = GetTrimmedAverage(sortedArray)
	return minValue, average, maxValue
end

local function CalculateItem(id, item)
	local minValue, averageValue, maxValue = GetAverageValues(item)
	local values = {}
	values.min = minValue
	values.avg = averageValue
	values.max = maxValue
	--print ("Item id=" .. id .. " min=" .. minValue .. " avg=" .. averageValue .. " max=" .. maxValue)
	HowMuch.derivedDb[id] = values				
end

local function RecalcDerivedDb()
	HowMuch.derivedDb = {}
end

local function ClearDerivedDbItem(itemId)
	HowMuch.derivedDb[itemId] = nil
end

local function GetTableForId(id)
	local ret = HowMuch.db[id]
	if ret == nil then
		ret = {}
		HowMuch.db[id] = ret
	end
	
	return ret
end

local function PurgeOldDates(id, latestDate)
	local idTable = GetTableForId(id)
	local lowestDate = latestDate
	local count = 0
	for dateString, _ in pairs(idTable) do
		count = count + 1
		if dateString < lowestDate then
			lowestDate = dateString
		end
	end
	
	if count > HowMuch.datesToKeep then
		idTable[lowestDate] = nil
	else
		return 0
	end
	
	return count - 1
end

local function GetDateTableForId(id, dateString)
	local idTable = GetTableForId(id)
	local ret = idTable[dateString]
	if ret == nil then
		ret = {}
		idTable[dateString] = ret
		while(PurgeOldDates(id, dateString) > HowMuch.datesToKeep) do
		end
		
		for thisDate, dateTable in pairs(idTable) do
			if thisDate ~= dateString then
				dateTable.ids = nil
			end
		end
	end
	
	return ret
end

local function GetPriceTableForId(id, dateString, price)
	local dateTable = GetDateTableForId(id, dateString)
	local ret = dateTable[price]
	if ret == nil then
		ret = {}
		ret.count = 0
		ret.ids = {}
		dateTable[price] = ret
	end
	
	return ret
end

local function RegisterPrice(dateString, auctionId, itemId, price, stacks)
	local unitPrice = math.floor(price / stacks)
	local priceTable = GetPriceTableForId(itemId, dateString, unitPrice)
	if priceTable.ids[auctionId] == nil then
		priceTable.ids[auctionId] = 1
		priceTable.count = priceTable.count + 1
	end
end

local function AuctionDetailUpdateHandler()
	if HowMuch.auctions == nil then
		if HowMuch.AuctionDetailStopCallback ~= nil then
			HowMuch.AuctionDetailStopCallback()
		end
	else
		local auctionIds = {}
		local count = 1
		for id, index in pairs (HowMuch.auctions) do
			auctionIds[id] = count
			count = count + 1
			if count > HowMuch.batchSize then
				break
			end
		end
		
		for id, index in pairs (auctionIds) do
			HowMuch.auctions[id] = nil
		end
		
		local details = Inspect.Auction.Detail(auctionIds)
		local dateString = os.date("%Y%m%d")
		for _, detail in pairs(details) do
			local stacks = detail.itemStack
			if stacks == nil then
				stacks = 1
			end
			
			if detail.buyout ~= nil then
				RegisterPrice(dateString, detail.id, detail.itemType, detail.buyout, stacks)
				ClearDerivedDbItem(detail.itemType)
			end
		end
		
		HowMuch.auctionsProcessed = HowMuch.auctionsProcessed + count
		SetPhaseDialog(5, HowMuch.auctionsProcessed, HowMuch.totalAuctions)
		
		if next(HowMuch.auctions) == nil then
			HowMuch.auctions = nil
		end
	end
end

local function StopAuctionDetailHandler()
	Command.Event.Detach(Event.System.Update.Begin, AuctionDetailUpdateHandler, "HowMuch.AuctionDetailUpdateHandler")
	HowMuch.AuctionDetailStopCallback = nil
	RecalcDerivedDb()
	HowMuch.window:SetVisible(false)
end

local function StartAuctionDetailHandler()
	HowMuch.AuctionDetailStopCallback = StopAuctionDetailHandler
	SetPhaseDialog(4)
	HowMuch.auctionsProcessed = 0
	Command.Event.Attach(Event.System.Update.Begin, AuctionDetailUpdateHandler, "HowMuch.AuctionDetailUpdateHandler")
end

local function ScanEventReceived(h, auctionType, auctions)
	local count = 0
	for _, index in pairs(auctions) do
		count = count + 1
	end
	
	HowMuch.auctions = auctions
	HowMuch.totalAuctions = count
	SetPhaseDialog(3, count)
	StartAuctionDetailHandler()
end

local function StartScan()
	if Inspect.Interaction("auction") == true then
		if Inspect.Queue.Status("auctionfullscan") ~= false then
			SetPhaseDialog(1)
			HowMuch.warningTooltip:Hide(HowMuch.button)
			HowMuch.window:SetVisible(true)
			local scanArgs = {}
			scanArgs.type = "search"
			--scanArgs.index = 1
			Command.Auction.Scan(scanArgs)
			HowMuch.dbRefreshDate = os.time()
			SetPhaseDialog(2)
		end
	end
end

local function TooltipEventReceived(h, tooltipType, itemId)
	if UI.Native.Tooltip:GetLoaded() ~= true then
		HowMuch.tooltip:Hide()
	elseif tooltipType == nil and HowMuch.tooltipShowFrame ~= nil then
		if HowMuch.tooltipShowFrame == Inspect.Time.Frame() then
			HowMuch.tooltip:Show()
		end
	end
	
	if tooltipType == "item" then
		local details = Inspect.Item.Detail(itemId)
		if details ~= nil and details.bound ~= true and details.type ~= nil then
			if HowMuch.derivedDb ~= nil and HowMuch.db[details.type] ~= nil then
				if HowMuch.derivedDb[details.type] == nil then
					CalculateItem(details.type, HowMuch.db[details.type])
				end
				
				local itemValues = HowMuch.derivedDb[details.type]
				local stacks = 1
				if details.stack ~= nil then
					stacks = details.stack
				end
				
				HowMuch.tooltip:SetValues(itemValues.avg * stacks, itemValues.min * stacks, itemValues.max * stacks)
				HowMuch.tooltipShowFrame = Inspect.Time.Frame()
				HowMuch.tooltip:Show()
			end
		end
	elseif tooltipType == "itemtype" then
		if itemId ~= nil then
			if HowMuch.derivedDb ~= nil and HowMuch.db[itemId] ~= nil then
				if HowMuch.derivedDb[itemId] == nil then
					CalculateItem(itemId, HowMuch.db[itemId])
				end
				
				local itemValues = HowMuch.derivedDb[itemId]
				HowMuch.tooltip:SetValues(itemValues.avg, itemValues.min, itemValues.max)
				HowMuch.tooltipShowFrame = Inspect.Time.Frame()
				HowMuch.tooltip:Show()
			end
		end
	end
end

local function ShowWarningTooltip()
	local oldestDate = os.time() - HowMuch.datesToKeep * 24 * 60 * 60
	if HowMuch.dbRefreshDate == nil or HowMuch.dbRefreshDate < oldestDate then
		local warningString = "HowMuch db out of date.\nPress spanner to refresh"
		if HowMuch.dbRefreshDate == nil then
			warningString = "HowMuch db not setup.\nPress spanner to scan prices"
		end
		
		HowMuch.warningTooltip:Show(HowMuch.button, warningString, "TOPRIGHT")
	else
		HowMuch.warningTooltip:Hide(HowMuch.button)
	end
end

local function InteractionEventReceived(h, interaction, state)
	if interaction == "auction" then
		if state == true then
			HowMuch.button:SetLayer(UI.Native.Auction:GetLayer() + 1)
			HowMuch.button:SetVisible(true)
			ShowWarningTooltip()
		else
			HowMuch.button:SetVisible(false)
			HowMuch.window:SetVisible(false)
		end
	end
end

local function SaveVariables(h, addon)
	if addon == HowMuch.name then
		-- now copy saved group to settings so that they can be preserved on logout
		HowMuch_SavedVariables = {}
		HowMuch_SavedVariables.version = HowMuch.version
		HowMuch_SavedVariables.lang = HowMuch.lang
		HowMuch_SavedVariables.db = HowMuch.db
		HowMuch_SavedVariables.dbRefreshDate = HowMuch.dbRefreshDate
	end
end

local function RegisterPostStartupEvents()
	--Command.Event.Attach(Command.Slash.Register(HowMuch.slashName), SlashHandler, "HowMuch.SlashHandler")
	Command.Event.Attach(Event.Addon.SavedVariables.Save.Begin, SaveVariables, "HowMuch.SaveVariables")
	Command.Event.Attach(Event.Auction.Scan, ScanEventReceived, "HowMuch.ScanEventReceived")
	Command.Event.Attach(Event.Tooltip, TooltipEventReceived, "HowMuch.TooltipEventReceived")
	Command.Event.Attach(Event.Interaction, InteractionEventReceived, "HowMuch.InteractionEventReceived")
end

local function Create()
	local margin = 5
	local parent = UI.CreateContext(HowMuch.name .. "Context")
	parent:SetStrata("tutorial")
	local HowMuchWindow = UI.CreateFrame("SimpleWindow", "HowMuchWindow", parent)
	HowMuchWindow:SetVisible(false)
	HowMuchWindow:SetCloseButtonVisible(true)
	HowMuchWindow:SetTitle("HowMuch")
	HowMuchWindow:SetPoint("TOPLEFT", UI.Native.Auction, "TOPLEFT", 10, 10)
	HowMuchWindow:SetWidth(300)
	HowMuchWindow:SetHeight(350)
	HowMuchWindow.text1 = UI.CreateFrame("Text", "HowMuchWindowText1", HowMuchWindow)
	HowMuchWindow.text1:SetPoint("TOPCENTER", HowMuchWindow, "TOPCENTER", 0, 50)
	HowMuchWindow.text2 = UI.CreateFrame("Text", "HowMuchWindowText2", HowMuchWindow)
	HowMuchWindow.text2:SetPoint("TOPCENTER", HowMuchWindow.text1, "TOPCENTER", 0, 20)
	HowMuchWindow.text3 = UI.CreateFrame("Text", "HowMuchWindowText3", HowMuchWindow)
	HowMuchWindow.text3:SetPoint("TOPCENTER", HowMuchWindow.text2, "TOPCENTER", 0, 20)
	HowMuchWindow.text4 = UI.CreateFrame("Text", "HowMuchWindowText4", HowMuchWindow)
	HowMuchWindow.text4:SetPoint("TOPCENTER", HowMuchWindow.text3, "TOPCENTER", 0, 20)
	HowMuchWindow.text5 = UI.CreateFrame("Text", "HowMuchWindowText5", HowMuchWindow)
	HowMuchWindow.text5:SetPoint("TOPCENTER", HowMuchWindow.text4, "TOPCENTER", 0, 20)
	HowMuchWindow.text6 = UI.CreateFrame("Text", "HowMuchWindowText6", HowMuchWindow)
	HowMuchWindow.text6:SetPoint("TOPCENTER", HowMuchWindow.text5, "TOPCENTER", 0, 20)
	HowMuch.window = HowMuchWindow
	
	local button = UI.CreateFrame("Texture", "HowMuchButton", parent)
	button:SetVisible(false)
	button:SetTexture(HowMuch.name, "SettingsButton32.png")
	button:SetPoint("TOPLEFT", UI.Native.Auction, "TOPLEFT", 18, 18)
	HowMuch.button = button
	local settingsButtonGlow = UI.CreateFrame("Texture", "HowMuchButtonGlow", button)
	settingsButtonGlow:SetTexture(HowMuch.name, "SettingsButtonGlow32.png")
	settingsButtonGlow:SetAllPoints(button)
	settingsButtonGlow:SetVisible(false)
	button:SetMouseMasking("limited")
	button.Event.MouseIn = function() settingsButtonGlow:SetVisible(true) end
	button.Event.MouseOut = function() settingsButtonGlow:SetVisible(false) end

	HowMuch.warningTooltip = UI.CreateFrame("SimpleTooltip", "HowMuchWarningToolip", button)
	
	local tooltipParent = UI.CreateContext(HowMuch.name .. "TooltipContext")
	tooltipParent:SetStrata("tooltip")
	HowMuch.tooltip = HowMuch.Tooltip.Create(HowMuch.name .. "Tooltip", tooltipParent)
end

local function Initialise(h, addon)
	if addon == HowMuch.name then
		HowMuch.startup = true
		Create()
	end
end

local function Startup()
	HowMuch.startCount = HowMuch.startCount + 1
	if HowMuch.startCount % 50 == 0 then
		if HowMuch.startup == true then
			HowMuch.startup = false

			RegisterPostStartupEvents()

			Command.Event.Detach(Event.System.Update.Begin, Startup, "HowMuch.Startup")
			
			RecalcDerivedDb()
			
			HowMuch.button:EventAttach(Event.UI.Input.Mouse.Left.Down, StartScan, "HowMuchButton.MouseClick")
		end
	end
end

local function GetSavedValue(value, default)
	if value ~= nil then
		return value
	else
		return default
	end
end

local function LoadVariables(h, addon)
	if addon == HowMuch.name then
		-- now that variables are loaded and saved positions restored we can create frame
		if HowMuch_SavedVariables then
			HowMuch.lang = GetSavedValue(HowMuch_SavedVariables.lang, "English")
			HowMuch.db = GetSavedValue(HowMuch_SavedVariables.db, {})
			HowMuch.dbRefreshDate = GetSavedValue(HowMuch_SavedVariables.dbRefreshDate, nil)
		end
		
		--HowMuch.SetLocale()
	end	
end

--
-- Register events
--
Command.Event.Attach(Event.Addon.Load.End, Initialise, "HowMuch.Initialise")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, LoadVariables, "HowMuch.LoadVariables")
Command.Event.Attach(Event.System.Update.Begin, Startup, "HowMuch.Startup")
